class Commands:

    """A list of IRC commands, which are used in `IRCConnection`.

    This list of commands is based on:
        https://en.wikipedia.org/wiki/List_of_Internet_Relay_Chat_commands
    """

    ADMIN = 'ADMIN'
    AWAY = 'AWAY'
    CNOTICE = 'CNOTICE'
    CPRIVMSG = 'CPRIVMSG'
    CONNECT = 'CONNECT'
    DIE = 'DIE'
    ENCAP = 'ENCAP'
    ERROR = 'ERROR'
    HELP = 'HELP'
    INFO = 'INFO'
    INVITE = 'INVITE'
    ISON = 'ISON'
    JOIN = 'JOIN'
    KICK = 'KICK'
    KILL = 'KILL'
    KNOCK = 'KNOCK'
    LINKS = 'LINKS'
    LIST = 'LIST'
    LUSERS = 'LUSERS'
    MODE = 'MODE'
    MOTD = 'MOTD'
    NAMES = 'NAMES'
    NAMESX = 'NAMESX'
    NICK = 'NICK'
    NOTICE = 'NOTICE'
    OPER = 'OPER'
    PART = 'PART'
    PASS = 'PASS'
    PING = 'PING'
    PONG = 'PONG'
    PRIVMSG = 'PRIVMSG'
    QUIT = 'QUIT'
    REHASH = 'REHASH'
    RESTART = 'RESTART'
    RULES = 'RULES'
    SERVER = 'SERVER'
    SERVICE = 'SERVICE'
    SERVLIST = 'SERVLIST'
    SQUERY = 'SQUERY'
    SQUIT = 'SQUIT'
    SETNAME = 'SETNAME'
    SILENCE = 'SILENCE'
    STATS = 'STATS'
    SUMMON = 'SUMMON'
    TIME = 'TIME'
    TOPIC = 'TOPIC'
    TRACE = 'TRACE'
    UHNAMES = 'UHNAMES'
    USER = 'USER'
    USERHOST = 'USERHOST'
    USERIP = 'USERIP'
    USERS = 'USERS'
    VERSION = 'VERSION'
    WALLOPS = 'WALLOPS'
    WATCH = 'WATCH'
    WHO = 'WHO'
    WHOIS = 'WHOIS'
    WHOWAS = 'WHOWAS'
